{ File   : BG2BMP.dpr
  Author : Deniz Oezmen
  Created: 2007-01-14
  Changed: 2007-01-14

  The Last Express BG converter
}
program XMG2BMP;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils, SaveBMP3;

type
  TBGHeader = packed record
                Unknown1,
                Unknown2: Longword;  // positions on screen?
                Width,
                Height,
                RSize,
                BSize,
                GSize   : Longint
              end;

  TBuffer   = array of Byte;

var
  BGFile: file;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenBGFile(var BGFile: file; BGName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(BGFile, BGName);

  {$I-}
  Reset(BGFile, 1);
  {$I+}
  OpenBGFile := (IOResult = 0)
end;

function ReadHeader(out BGHeader: TBGHeader): Boolean;
var
  BytesRead: Longword;
begin
  BlockRead(BGFile, BGHeader, SizeOf(TBGHeader), BytesRead);
  ReadHeader := (SizeOf(TBGHeader) = BytesRead)
end;

procedure DecompressStream(const InBuffer: TBuffer; var OutBuffer: TBuffer);
var
  i,
  InPos,
  HisPos,
  OutPos: Longint;
  Len,
  Data  : Byte;
begin
  InPos := 0;
  OutPos := 0;

  while InPos <= High(InBuffer) do
  begin
    if InBuffer[InPos] < $80 then
    begin
      // direct decompression
      Len := InBuffer[InPos] shr 5 + 1;
      Data := InBuffer[InPos] shl 3;
      Inc(InPos);

      for i := 1 to Len do
      begin
        if OutPos > High(OutBuffer) then
          Exit;

        OutBuffer[OutPos] := Data;
        Inc(OutPos)
      end
    end
    else
    begin
      // buffer back reference, 4096 byte window
      if InPos = High(InBuffer) then
        Exit;

      Len := (InBuffer[InPos] and $7f) shr 4 + 3;
      HisPos := OutPos - 4096 +
        (InBuffer[InPos] and $0f) shl 8 + InBuffer[InPos + 1];
      Inc(InPos, 2);

      if HisPos < 0 then
        Exit;

      for i := HisPos to HisPos + Len - 1 do
      begin
        if OutPos > High(OutBuffer) then
          Exit;
        if i > High(OutBuffer) then
          Exit;

        OutBuffer[OutPos] := OutBuffer[i];
        Inc(OutPos)
      end
    end
  end;
end;

procedure Convert(BGName: string; const BGHeader: TBGHeader);
var
  i,
  x, y : Longint;
  BMP  : TBMPFile;
  RData,
  GData,
  BData,
  ROut,
  GOut,
  BOut : TBuffer;
begin
  with BGHeader do
    WriteLn(Format(' (%d x %d)', [Width, Height]));

  // get compressed colour streams
  SetLength(RData, BGHeader.RSize);
  SetLength(BData, BGHeader.BSize);
  SetLength(GData, BGHeader.GSize);

  BlockRead(BGFile, RData[0], BGHeader.RSize);
  BlockRead(BGFile, BData[0], BGHeader.BSize);
  BlockRead(BGFile, GData[0], BGHeader.GSize);

  // decompress streams
  SetLength(ROut, BGHeader.Width * BGHeader.Height);
  SetLength(GOut, BGHeader.Width * BGHeader.Height);
  SetLength(BOut, BGHeader.Width * BGHeader.Height);

  DecompressStream(RData, ROut);
  DecompressStream(BData, BOut);
  DecompressStream(GData, GOut);

  // write bitmap
  BGName := Prefix(BGName) + '.bmp';
  OpenBMP(BMP, BGName, BGHeader.Width, BGHeader.Height);

  // not very efficient, but what the hell ...
  for y := 0 to BGHeader.Height - 1 do
    for x := 0 to BGHeader.Width - 1 do
    begin
      i := y * BGHeader.Width + x;
      WritePixelToBMP(BMP, x, y, ROut[i], GOut[i], BOut[i]);
    end;

  CloseBMP(BMP);

  Finalize(RData);
  Finalize(GData);
  Finalize(BData);
  Finalize(ROut);
  Finalize(GOut);
  Finalize(BOut)
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  BGName   : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
  BGHeader : TBGHeader;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara
  if FindFirst(PFilePara, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      BGName := Dir + FileName;

      Write(BGName);

      // if this file is accessible, analyze it
      if OpenBGFile(BGFile, BGName) then
      begin
        if ReadHeader(BGHeader) then
          Convert(BGName, BGHeader)
        else
          WriteLn(' is not a valid BG file');
        Close(BGFile)
      end
      else
        Writeln(BGName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found')
end;

begin
  if ParamCount >= 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.

