{ File   : SaveBMP3.pas
  Author : Deniz Oezmen, based on work by Christian Klaessen and Tobias Post
  Created: 1999-12-11
  Changed: 2005-11-12

  Library providing various functions to export data as a standard Windows
  bitmap file.
}
unit SaveBMP3;

interface

  const
    Version = $0401;
    Build   = $0000;
    Author  = 'T. Post, C. Klaessen and D. Oezmen, 1999-2005';

  type
    TColor = record
               B, G, R, A: Byte
             end;

    TColor3 = record
                B, G, R: Byte;
              end;

    TPalette = packed array[0..255] of TColor;

    TPalette3 = packed array[0..255] of TColor3;


    TBitmapFileHeader = packed record
                         bfType     : Word;    {UINT  : Erkennungsstring = 'BM'}
                         bfSize     : Longint; {DWORD : Gre der Bitmapdatei}
                         bfReserved1,          {UINT  : Reserviert1 = 0}
                         bfReserved2: Word;    {UINT  : Reserviert2 = 0}
                         bfOffBits  : Longint  {DWORD : Beginn (Offset) der Bilddaten}
                        end;

    TBitmapInfoHeader = packed record
                          biSize,                   {DWORD : Gre des BMIH}
                          biWidth,                  {LONG  : Breite des Bildes}
                          biHeight       : Longint; {LONG  : Hhe des Bildes}
                          biPlanes,                 {WORD  : Bildebenen? = 1}
                          biBitCount     : Word;    {WORD  : Bits pro Pixel = 1/4/8/24}
                          biCompression,            {DWORD : Komprimierung = 0/1}
                          biSizeImage,              {DWORD : Gre der Bilddaten}
                          biXPelsPerMeter,          {LONG  : X-Pixelanzahl pro Meter}
                          biYPelsPerMeter,          {LONG  : Y-Pixelanzahl pro Meter}
                          biClrUsed,                {DWORD : Farbanzahl in Farbpalette}
                          biClrImportant : Longint  {DWORD : Bits Pro Pixel des Grafiktreibers}
                        END;

     TBMPFile         = packed record
                          BMPFile   : file;
                          Buffer    : array of Byte;
                          FileHeader: TBitmapFileHeader;
                          InfoHeader: TBitmapInfoHeader
                        end;

  procedure OpenBMP(var BMP: TBMPFile; BMPName: string; XRes, YRes: Longint);
  procedure WritePixelToBMP(var BMP: TBMPFile; x, y: Longint; R, G, B: Byte);
  procedure CloseBMP(BMP: TBMPFile);
  procedure Dump8BitBMP(BMPName: string; XRes, YRes: Longint; Palette: TPalette;
    Content: Pointer); overload;
  procedure Dump8BitBMP(BMPName: string; XRes, YRes: Longint;
    Palette3: TPalette3; Content: Pointer); overload;

implementation

procedure PrepareBMP(var BMP: TBMPFile; BMPName: string; XRes, YRes: Longint;
  Bits: Word);
var
  ScanX,
  ScanY     : Longint;
begin
  Assign(BMP.BMPFile, BMPName);
  Rewrite(BMP.BMPFile, 1);

  ScanX := XRes;
  ScanY := YRes;

  ScanX := (ScanX div 4 + Byte(ScanX mod 4 > 0)) * 4;

  with BMP.FileHeader do
  begin
    bfType := Ord('B') + Ord('M') * 256;
    bfSize := SizeOf(BMP.FileHeader) + SizeOf(BMP.InfoHeader) +
      3 * ScanX * ScanY;
    bfReserved1 := $00;
    bfReserved2 := $00;
    bfOffBits := SizeOf(BMP.FileHeader) + SizeOf(BMP.InfoHeader)
  end;

  with BMP.InfoHeader do
  begin
    biSize := SizeOf(BMP.InfoHeader);
    biWidth := XRes;
    biHeight := YRes;
    biPlanes := $01;
    biBitCount := Bits;
    biCompression := $0000;
    biSizeImage := 3 * ScanX * ScanY;
    biXPelsPerMeter := $0000;
    biYPelsPerMeter := $0000;
    biClrUsed := $0000;
    biClrImportant := $0000
  end;

  BlockWrite(BMP.BMPFile, BMP.FileHeader, SizeOf(BMP.FileHeader));
  BlockWrite(BMP.BMPFile, BMP.InfoHeader, SizeOf(BMP.InfoHeader))
end;

procedure FillBMP(var BMP: TBMPFile);
begin
  SetLength(BMP.Buffer, BMP.InfoHeader.biSizeImage);
  FillChar(BMP.Buffer[0], BMP.InfoHeader.biSizeImage, 0)
end;

procedure OpenBMP(var BMP: TBMPFile; BMPName: string; XRes, YRes: Longint);
begin
  PrepareBMP(BMP, BMPName, XRes, YRes, 24);
  FillBMP(BMP)
end;

procedure WritePixelToBMP(var BMP: TBMPFile; x, y: Longint; R, G, B: Byte);
var
  XRes,
  YRes,
  p   : Longint;
begin
  XRes := Longint(BMP.InfoHeader.biWidth);
  YRes := Longint(BMP.InfoHeader.biHeight);

  if (x >= XRes) or (y >= YRes) then
    Exit;

  p := (YRes - y - 1) * 3 * XRes + (YRes - y - 1) * (XRes mod 4) + 3 * x;
  BMP.Buffer[p    ] := B;
  BMP.Buffer[p + 1] := G;
  BMP.Buffer[p + 2] := R
end;

procedure CloseBMP(BMP : TBMPFile);
begin
  BlockWrite(BMP.BMPFile, BMP.Buffer[0], BMP.InfoHeader.biSizeImage);
  Close(BMP.BMPFile)
end;

procedure Prepare8BitBMP(var BMP: TBMPFile; BMPName: string;
  XRes, YRes: Longint; Bits: Word);
var
  ScanX,
  ScanY     : Longint;
begin
  Assign(BMP.BMPFile, BMPName);
  Rewrite(BMP.BMPFile, 1);

  ScanX := XRes;
  ScanY := YRes;

  ScanX := (ScanX div 4 + Byte(ScanX mod 4 > 0)) * 4;

  with BMP.FileHeader do
  begin
    bfType := Ord('B') + Ord('M') * 256;
    bfSize := SizeOf(BMP.FileHeader) + SizeOf(BMP.InfoHeader) +
      SizeOf(TPalette) + ScanX * ScanY;
    bfReserved1 := $00;
    bfReserved2 := $00;
    bfOffBits := SizeOf(BMP.FileHeader) + SizeOf(BMP.InfoHeader) +
      SizeOf(TPalette);
  end;

  with BMP.InfoHeader do
  begin
    biSize := SizeOf(BMP.InfoHeader);
    biWidth := XRes;
    biHeight := YRes;
    biPlanes := $01;
    biBitCount := Bits;
    biCompression := $0000;
    biSizeImage := 3 * ScanX * ScanY;
    biXPelsPerMeter := $0000;
    biYPelsPerMeter := $0000;
    biClrUsed := $0000;
    biClrImportant := $0000
  end;

  BlockWrite(BMP.BMPFile, BMP.FileHeader, SizeOf(BMP.FileHeader));
  BlockWrite(BMP.BMPFile, BMP.InfoHeader, SizeOf(BMP.InfoHeader))
end;

procedure Dump8BitBMP(BMPName: string; XRes, YRes: Longint; Palette: TPalette;
  Content: Pointer);
var
  BMP  : TBMPFile;
  Line : Pointer;
  y,
  Width: Word;
begin
  Prepare8BitBMP(BMP, BMPName, XRes, YRes, 8);

  BlockWrite(BMP.BMPFile, Palette, SizeOf(Palette));

  Width := BMP.InfoHeader.biWidth;
  Width := (Width div 4 + Byte(Width mod 4 > 0)) * 4;

  GetMem(Line, Width);
  FillChar(Line^, Width, 0);

  for y := BMP.InfoHeader.biHeight - 1 downto 0 do
  begin
    Move(Pointer(Longint(Content) + (y * BMP.InfoHeader.biWidth))^,
      Line^, BMP.InfoHeader.biWidth);
    BlockWrite(BMP.BMPFile, Line^, Width)
  end;

  FreeMem(Line, Width);

  CloseBMP(BMP)
end;

procedure ConvertPalette3toPalette(Palette3: TPalette3; var Palette: TPalette);
var
  i: Byte;
begin
  for i := 0 to 255 do
  begin
    Palette[i].B := Palette3[i].B;
    Palette[i].G := Palette3[i].G;
    Palette[i].R := Palette3[i].R;
    Palette[i].A := 0
  end
end;

procedure Dump8BitBMP(BMPName: string; XRes, YRes: Longint; Palette3: TPalette3;
  Content: Pointer);
var
  Palette: TPalette;
begin
  ConvertPalette3ToPalette(Palette3, Palette);
  Dump8BitBMP(BMPName, XRes, YRes, Palette, Content)
end;

begin
end.
